use anyhow::{bail, Context, Result};
use cgmath::{Deg, Point2};
use std::collections::{HashMap, HashSet};

use crate::physics::{Pfloat, Polar};

/// Toml file syntax
pub(in crate::content) mod toml {
	use super::Pfloat;
	use serde::Deserialize;
	use std::collections::HashMap;

	#[derive(Debug, Deserialize)]
	pub struct SystemRoot {
		pub system: System,
		pub object: HashMap<String, Object>,
	}

	#[derive(Debug, Deserialize)]
	pub struct System {
		pub name: String,
	}

	#[derive(Debug, Deserialize)]
	pub struct Object {
		pub sprite: String,
		pub position: Position,

		pub size: Pfloat,
		pub parallax: Pfloat,

		pub radius: Option<Pfloat>,
		pub angle: Option<Pfloat>,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Position {
		Polar(PolarCoords),
		Cartesian(Coordinates),
	}

	#[derive(Debug, Deserialize)]
	pub struct PolarCoords {
		pub center: Coordinates,
		pub radius: Pfloat,
		pub angle: Pfloat,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Coordinates {
		Label(String),
		Coords([Pfloat; 2]),
	}

	impl ToString for Coordinates {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_owned(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}
}

#[derive(Debug)]
pub struct System {
	pub name: String,
	pub objects: Vec<Object>,
}

#[derive(Debug)]
pub struct Object {
	pub sprite: String,
	pub position: Point2<f32>,
	pub size: Pfloat,
	pub parallax: Pfloat,
	pub angle: Deg<Pfloat>,
}

fn resolve_coordinates(
	objects: &HashMap<String, toml::Object>,
	cor: &toml::Coordinates,
	mut cycle_detector: HashSet<String>,
) -> Result<Point2<f32>> {
	match cor {
		toml::Coordinates::Coords(c) => Ok((*c).into()),
		toml::Coordinates::Label(l) => {
			if cycle_detector.contains(l) {
				bail!(
					"Found coordinate cycle: `{}`",
					cycle_detector.iter().fold(String::new(), |sum, a| {
						if sum.is_empty() {
							a.to_string()
						} else {
							sum + " -> " + a
						}
					})
				);
			}
			cycle_detector.insert(l.to_owned());

			let p = match objects.get(l) {
				Some(p) => p,
				None => bail!("Could not resolve coordinate label `{l}`"),
			};
			Ok(resolve_position(&objects, &p, cycle_detector)
				.with_context(|| format!("In object {:#?}", l))?)
		}
	}
}

fn resolve_position(
	objects: &HashMap<String, toml::Object>,
	obj: &toml::Object,
	cycle_detector: HashSet<String>,
) -> Result<Point2<f32>> {
	match &obj.position {
		toml::Position::Cartesian(c) => Ok(resolve_coordinates(objects, c, cycle_detector)?),
		toml::Position::Polar(p) => Ok(Polar {
			center: resolve_coordinates(&objects, &p.center, cycle_detector)?,
			radius: p.radius,
			angle: Deg(p.angle),
		}
		.to_cartesian()),
	}
}

impl System {
	pub fn parse(value: toml::SystemRoot) -> Result<Self> {
		let mut objects = Vec::new();

		for (label, obj) in &value.object {
			let mut cycle_detector = HashSet::new();
			cycle_detector.insert(label.to_owned());

			objects.push(Object {
				sprite: obj.sprite.clone(),
				position: resolve_position(&value.object, &obj, cycle_detector)
					.with_context(|| format!("In object {:#?}", label))?,
				size: obj.size,
				parallax: obj.parallax,
				angle: Deg(obj.angle.unwrap_or(0.0)),
			});
		}

		return Ok(Self {
			name: value.system.name.clone(),
			objects,
		});
	}
}
